/******************************************************************************
** Copyright (c) 2013-2016 Intel Corporation All Rights Reserved
**
** Licensed under the Apache License, Version 2.0 (the "License"); you may not
** use this file except in compliance with the License.
**
** You may obtain a copy of the License at
**     http://www.apache.org/licenses/LICENSE-2.0
**
** Unless required by applicable law or agreed to in writing, software
** distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
** WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
**
** See the License for the specific language governing permissions and
** limitations under the License.
**
******************************************************************************/

#define ESIF_TRACE_ID	ESIF_TRACEMODULE_EVENT

#include "esif_uf.h"		/* Upper Framework */
#include "esif_ws_socket.h"
#include "esif_ws_server.h"
#include "esif_uf_fpc.h"
#include "esif_uf_domain.h"
#include "esif_participant.h"
#include "esif_uf_eventmgr.h"
#include "esif_pm.h"

static eEsifError ESIF_CALLCONV ControlActionCallback(
	void *contextPtr,
	UInt8 participantId,
	UInt16 domainId,
	EsifFpcEventPtr fpcEventPtr,
	EsifDataPtr eventDataPtr)
{
	EsifCapabilityDataPtr capabilityDataPtr = NULL;
	eEsifError rc = ESIF_OK;

	UNREFERENCED_PARAMETER(contextPtr);
	// Sanity checks
	if (ESIF_EVENT_DPTF_PARTICIPANT_CONTROL_ACTION != fpcEventPtr->esif_event) {
		rc = ESIF_E_NOT_SUPPORTED;
		goto exit;
	}

	if (NULL == eventDataPtr) {
		rc = ESIF_E_PARAMETER_IS_NULL;
		goto exit;
	}

	if (eventDataPtr->data_len < sizeof(*capabilityDataPtr)) {
		rc = ESIF_E_PARAMETER_IS_OUT_OF_BOUNDS;
		goto exit;
	}

	capabilityDataPtr = (EsifCapabilityDataPtr)eventDataPtr->buf_ptr;
	if (NULL == capabilityDataPtr) {
		rc = ESIF_E_PARAMETER_IS_NULL;
		goto exit;
	}

	ESIF_TRACE_DEBUG("ESIF_EVENT_DPTF_PARTICIPANT_CONTROL_ACTION: Participant ID: %d, domain ID: 0x%x\n",
		participantId, domainId);

	rc = esif_ws_broadcast_data_buffer((u8 *)capabilityDataPtr, capabilityDataPtr->size);

exit:
	return rc;
}

static void EnableControlActionReportingForDomain(EsifUpDomainPtr domainPtr)
{
	UInt32 capMask = EsifUp_GetDomainCapabilityMask(domainPtr);
	EsifData capData = { ESIF_DATA_UINT32, &capMask, sizeof(capMask), sizeof(capMask) };

	EsifEventMgr_SignalEvent(domainPtr->participantId,
		domainPtr->domain,
		ESIF_EVENT_DPTF_PARTICIPANT_ACTIVITY_LOGGING_ENABLED,
		&capData
	);
}

static Bool ShouldEnableControlActionReportingForDomain(EsifUpDomainPtr domainPtr)
{
	Bool rc = ESIF_FALSE;

	ESIF_ASSERT(domainPtr != NULL);

	switch (domainPtr->domainType) {
	case ESIF_DOMAIN_TYPE_PROCESSOR:
	case ESIF_DOMAIN_TYPE_GRAPHICS:
	case ESIF_DOMAIN_TYPE_FAN:
	case ESIF_DOMAIN_TYPE_WIRELESS:
	case ESIF_DOMAIN_TYPE_MULTIFUNCTION:
	case ESIF_DOMAIN_TYPE_DISPLAY:
	case ESIF_DOMAIN_TYPE_BATTERYCHARGER:
	case ESIF_DOMAIN_TYPE_WWAN:
	case ESIF_DOMAIN_TYPE_POWER:
		rc = ESIF_TRUE;
	}

	return rc;
}

static void EnableControlActionReporting(EsifUpPtr upPtr)
{
	UpDomainIterator udIter = { 0 };
	EsifUpDomainPtr domainPtr = NULL;
	eEsifError iterRc = ESIF_OK;

	ESIF_ASSERT(upPtr != NULL);
	
	iterRc = EsifUpDomain_InitIterator(&udIter, upPtr);
	if (ESIF_OK != iterRc) goto exit;

	iterRc = EsifUpDomain_GetNextUd(&udIter, &domainPtr);
	while (ESIF_OK == iterRc) {
		// Found a valid domain
		if (ShouldEnableControlActionReportingForDomain(domainPtr)) {
			EnableControlActionReportingForDomain(domainPtr);
		}

		iterRc = EsifUpDomain_GetNextUd(&udIter, &domainPtr);
	}

	if (ESIF_E_ITERATION_DONE != iterRc) {
		EsifUp_PutRef(upPtr);
	}

exit:
	return;
}


eEsifError EsifControlActionBroadcastEnableParticipant(EsifUpPtr upPtr)
{
	eEsifError rc = ESIF_OK;

	if (NULL == upPtr) {
		ESIF_TRACE_ERROR("Invalid Upper Participant");
		rc = ESIF_E_PARAMETER_IS_NULL;
		goto exit;
	}

	// Register for the DPTF Participant Control Action Callback
	// All participants are going to share a single callback
	EsifEventMgr_RegisterEventByType(ESIF_EVENT_DPTF_PARTICIPANT_CONTROL_ACTION,
		EVENT_MGR_MATCH_ANY,
		EVENT_MGR_MATCH_ANY,
		ControlActionCallback,
		NULL
	);
		
	EnableControlActionReporting(upPtr);
exit:
	return rc;
}

/*****************************************************************************/
/*****************************************************************************/
/*****************************************************************************/
